# AUTOGENERATED! DO NOT EDIT! File to edit: 02_recursion.ipynb (unless otherwise specified).

__all__ = ['depth', 'flatten', 'unflatten', 'get', 'give', 'sort', 'pipe', 'maps']

# Cell
import typing
import numpy as np
from typing import Optional, Tuple, Dict, Callable, Union, Mapping, Sequence, Iterable
from functools import partial
import warnings
from .conversion import cast,convert
import collections

# Cell
def depth(obj : Union[dict,list]):
    '''
    Recursively sort the number of layers of a nested
    list or dictionary `x`.
    '''
    if type(obj) is dict and obj:
        return 1 + max(depth(obj[a]) for a in obj)
    if type(obj) is list and obj:
        return 1 + max(depth(a) for a in obj)
    return 0


# Cell
def flatten(obj : Union[dict,list], parent_key='',sep=','):
    '''
    Concatenate the nested input `obj` into an equivalent
    datastructure of depth 1. Uses the `parent_key` and `sep`
    arg to combine nested dictionary keys.
    https://stackoverflow.com/questions/6027558/flatten-nested-dictionaries-compressing-keys

    '''

    def flatten_list(obj):
        if obj == []:
            return obj
        if isinstance(obj[0], list):
            return flatten_list(obj[0]) + flatten_list(obj[1:])
        return obj[:1] + flatten_list(obj[1:])

    def flatten_dict(obj=obj,parent_key=parent_key,sep=sep):
        items = []
        for k, v in obj.items():
            new_key = parent_key + sep + str(k) if parent_key else str(k)
            if isinstance(v, collections.abc.MutableMapping):
                items.extend(flatten_dict(v, new_key, sep=sep).items())
            else:
                items.append((new_key, v))
        return dict(items)

    if type(obj) is dict:
            return flatten_dict(obj)
    elif type(obj) is list:
        return flatten_list(obj)


# Cell
def unflatten(d : dict, sep=","):
    '''
    Un-flatten a flattened nested dictionary `d` with
    concatenated key separation `sep`.
    https://gist.github.com/fmder/494aaa2dd6f8c428cede
    '''
    items = dict()
    for k, v in d.items():
        keys = k.split(sep)
        sub_items = items
        for ki in keys[:-1]:
            try:
                sub_items = sub_items[ki]
            except KeyError:
                sub_items[ki] = dict()
                sub_items = sub_items[ki]

        sub_items[keys[-1]] = v

    return items

# Cell
def get(obj : Union[object,dict,list,tuple,callable,np.ndarray],
        *args : Union[None,str,int,float,tuple,list,dict],
        retnone : bool = True,
        call : bool = False,
        **kwargs
       ) -> Union[object,dict,list,tuple,callable,int,float,set,np.ndarray]:
    '''
    Recursively accesses `obj` by the given ordered attributes/keys/indexes `args`.

    If `call`, the last accessed object will try to call any **kwargs.

    If `retnone`, returns None if
    the object can't be accessed by any of the args; else, returns `obj`.
    '''
    res=obj
    for attr in args:
        try: #list, arr, tuple w/index=attr, dict w/ key=attr
            res=res[attr]
        except:
            try: #class object w/ attr
                res=getattr(res,attr)
            except: #method or callable
                try:
                    res=res(attr)
                except:
                    pass

    if call:
        try:
            res=res(**kwargs)
        except:
            res=res

    if res is not obj:
        return res
    else:
        return (None if retnone else obj)

# Cell
def give(obj : Union[object,dict,list,tuple,np.ndarray],
        *args : Union[str,int,float,tuple,list,dict],
         **kwargs : object
       ) -> Union[object,dict,list,tuple,np.ndarray]:
    '''
    Assign `args` element of `obj` to a value. If no `kwargs`,
    assign the last item of `args`. If `kwargs`, assign those.
    '''

    if kwargs=={}:# no key-value pairs to set, use last ordered arg, assuming len(args)>1:
        if len(args)<2:
            print('Not enough args or kwargs - require access to elements of obj.')

        elif len(args)==2: #index by first, set value of second
            try:
                obj[args[0]]=args[1]
            except:
                try:
                    setattr(obj,args[0],args[1])
                except:
                    print(f"Could not access up to {args[:-1]} and/or apply {args[-1]}.")

        elif len(args)>2: #get up to -3, access by -2, set value with -1
            temp=get(obj,*args[:-2]) #need to access temporary variable
            try:
                temp[args[-2]]=args[-1]
            except:
                try:
                    setattr(temp,args[-2],args[-1])
                except:
                    print(f"Could not access up to {args[:-1]} and/or apply {args[-1]}.")

    else: #dict or obj-like
        if len(args)==0: #set key,value pairs or attrs
            try:
                for k,v in kwargs.items():
                    try:
                        obj[k]=v
                    except:
                        setattr(obj,k,v)
            except:
                print(f"Could not access {k} and/or apply {v}.")

        else:
            try:
                temp=get(obj,*args,retnone=False)
                for k,v in kwargs.items():
                    try:
                        temp[k]=v
                    except:
                        setattr(temp,k,v)
            except:
                print(f"Could not access {k} and/or apply {v}.")



# Cell
def sort(obj : Iterable,
         *args : Union[None,str,int,float,tuple,list,dict],
         by : Union[None,object,dict,list,tuple,callable,np.ndarray] = None,#lambda o: o,
         key : callable = lambda t: get(t,-1,retnone=False) if get(t,-1,retnone=False) else 0,#lambda t:t[-1] if depth(t)!=0 else t,
         sift : callable = lambda t: True,# if get(t,-1,retnone=False) else False,
         reverse : bool = True
        ) -> Union[None,int,float,list,tuple,str,dict,np.ndarray]:
    '''
    Recursively sorts `obj` `by` `args` using `key`.

    Treats `obj` as an iterator and wraps `by` around the elements of `obj`,

    before optionally wrapping any remaining `args`, and returning the evaluated

    tuples generated over `obj`. If `by` is None, searches for `args`.

    If both `by` and `args` is None, sorts over `obj` only : default behavior.

    `sift` defaults to keeping all elements, while

    `key` defaults to treating empty objects as having value 0.

    `reverse` sorts ascending by default.

    '''

    if by is not None:#if you're sorting over a different evaluation than the object elements

        sorting=filter(sift,[ (i, get( get( by, i, retnone=False), #evaluate inner function
                        *args,retnone=False) #and any remaining args
                  ) for i in obj]) #over iterable and store tuples then sort

    elif args is not (): #if you're sorting over the object and provide args

        sorting=filter(sift,[get(get(obj,i,retnone=False),*args,retnone=False) for i in obj])

    else: #otherwise you're just sorting the object

        sorting=filter(sift,obj)

    return sorted(sorting,key=key,reverse=reverse)


# Cell
def pipe(func,otype=None,ftype=None,cast=cast,*args,**kwargs):
    '''
    Pipelines the `func` to act on a later object.
    Returns a partially evaluated function over any `args` and `kwargs`.
    The object is casted to type `otype` before being evaluated.
    The output of the function is casted to `ftype`
    '''
    f=partial(func,*args,**kwargs)
    def line(obj):
        obj=cast(obj,otype)
        return cast(f(obj),ftype)
    return line


# Cell
def maps(obj,
             *funcs,
             depth=0,
             zipit=False,
             to=None,
             squeeze=True):
    '''
    Sequentially map `funcs` over the elements of `obj`, "o".
    The first `depth` number of funcs are mapped sequentially f(g(h(...(o)..)))=x.
    The remaining number of funcs are mapped separately (u(x),v(x),...).
    Use partial `funcs` to fill in all args but `obj` if other parameters needed.
    If `keys`, return tuples of the object elements "o" along with map outputs.
    '''
    if not funcs:
        return obj
    else:
        obj=obj if hasattr(obj,'__iter__') and type(obj)!='str' else [obj] #asiter(obj)
        r=[o for o in obj]
        [[give(r,i,get(f,r[i])) for f in funcs[:depth]] for i in range(len(r))]
        [give(r,i,[get(f,r[i]) for f in funcs[depth:]]) for i in range(len(r))]
        if squeeze:
            r=np.ndarray.tolist(np.squeeze(np.array(r,dtype=object)))
        if zipit:
            r=list(zip(obj,r))
        return cast(r,to)


# AUTOGENERATED! DO NOT EDIT! File to edit: 00_utils.ipynb (unless otherwise specified).

__all__ = ['timer', 'filesize', 'save', 'load', 'push', 'refresh', 'backup', 'figParams', 'forceAspect', 'matshow']

# Cell
#from typing import Optional, Tuple, Dict, Callable, Union, Mapping, Sequence
import time
import gzip
import pickle
import os

# Cell
def timer(func : callable) -> str:
    '''Decorator that reports the execution time.'''
    def wrap(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        print(func.__name__+' : '+f"Elapsed time: {end-start:0.4f} seconds")
        return result
    return wrap

# Cell
def filesize(fname : str ='data.gz',
             exp : int = 1e-6):
    '''
    Returns filesize in bytes*`exp`, which defaults to megabytes.
    '''
    prefix={1:'', 1e-3:'kilo', 1e-6:'mega', 1e-9:'giga'}
    size=float((os.stat(fname).st_size)*(exp if exp in prefix else 1))
    return f"{fname} is {size:.5} {prefix.get(exp) or ''}bytes"

# Cell
def save(data,fname='data.gz'):
    '''
    Saves `data` as `fname` using gzip and pickle.
    Maximizes speed and compression of objects.
    '''
    with gzip.open(fname, "wb") as f:
        pickle.dump(data,f,-1)
    print(filesize(fname))

# Cell
def load(fname='data.gz',delete=False):
    '''
    Loads `fname`, returning pickled `data`.
    '''
    with gzip.open(fname, "rb") as f:
        data=pickle.load(f)
    if delete is True:
        os.remove(fname)
    return data

# Cell
def push(branch='master',comment='auto'):
    "Pushes all current files to given branch with comment."
    os.system('git branch -M {}'.format(branch))
    os.system('git add -A') #stage all files
    os.system('git commit -m "{}"'.format(comment)) #commit all files
    os.system('git push -u origin {} --force'.format(branch))

def refresh(comment='auto',branch='master'):
    "Builds nbdev library and docs."
    os.system('nbdev_build_lib') #build core python module
    os.system('nbdev_install_git_hooks')
    os.system('nbdev_build_docs') #build code documentation
    push(comment=comment,branch=branch)

def backup(comment='Backup'):
    "Like `Push` but with branch set to `backup`."
    os.system('git branch -M backup')
    os.system('git add -A') #stage all files
    os.system('git commit -m "{}"'.format(comment)) #commit all files
    os.system('git push -u origin backup --force')

# Cell
import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np

def figParams(reset : bool = False,
              X : float = 3.5,
              Y : float = 3,
              hspace : float = 0.0,
              offset : float = -.4,
              font : str = 'Times New Roman',
              fontsize : int = 12,
              ticksize : int = 6,
              tickwidth : int = 2,
              linewidth : int = 2
             ):
    '''
    Changes the `rcParams` for plotting, with the option to `reset` to default.
    '''
    if reset:
        mpl.rcParams.update(mpl.rcParamsDefault)
    else:
        #figure font Times New Roman,Helvetica, Arial, Cambria, or Symbol
        mpl.rcParams['font.family'] = font
        mpl.rcParams['font.size'] = fontsize
        mpl.rcParams['axes.titlesize'] = fontsize
        mpl.rcParams['axes.linewidth'] = linewidth
        mpl.rcParams['axes.titley'] = offset


        mpl.rcParams['xtick.major.size'] = ticksize
        mpl.rcParams['xtick.major.width'] = tickwidth
        mpl.rcParams['xtick.minor.size'] = ticksize//2
        mpl.rcParams['xtick.minor.width'] = tickwidth//2
        mpl.rcParams['xtick.direction']='out'

        mpl.rcParams['ytick.major.size'] = ticksize
        mpl.rcParams['ytick.major.width'] = tickwidth
        mpl.rcParams['ytick.minor.size'] = ticksize//2
        mpl.rcParams['ytick.minor.width'] = tickwidth//2
        mpl.rcParams['ytick.direction']='out'

        mpl.rcParams['figure.figsize'] = [X,Y]
        mpl.rcParams['figure.subplot.hspace'] = hspace

def forceAspect(ax : mpl.axes,
                aspect : int = 1):
    '''
    Forces the aspect of the axes object `ax`.
    '''
    im = ax.get_images()
    extent =  im[0].get_extent()
    ax.set_aspect(abs((extent[1]-extent[0])/(extent[3]-extent[2]))/aspect)

def matshow(x : np.ndarray,
            aspect : int = 1,
            save : bool = True,
            fname : str = 'test'):
    '''
    Simplified image plot of matrix `x` with forced `aspect` that can save `fname` to `path`.
    '''
    fig,ax=plt.subplots()
    ax.matshow(x)
    forceAspect(ax,aspect)
    if save:
        plt.savefig(fname, dpi=600,transparent=False, bbox_inches='tight')


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"timer": "00_utils.ipynb",
         "filesize": "00_utils.ipynb",
         "save": "00_utils.ipynb",
         "load": "00_utils.ipynb",
         "push": "00_utils.ipynb",
         "refresh": "00_utils.ipynb",
         "backup": "00_utils.ipynb",
         "figParams": "00_utils.ipynb",
         "forceAspect": "00_utils.ipynb",
         "matshow": "00_utils.ipynb",
         "data": "01_conversion.ipynb",
         "trycast": "01_conversion.ipynb",
         "nonitr2itr": "01_conversion.ipynb",
         "itr2nonitr": "01_conversion.ipynb",
         "itr2itr": "01_conversion.ipynb",
         "Caster": "01_conversion.ipynb",
         "isiter": "01_conversion.ipynb",
         "cast": "01_conversion.ipynb",
         "typestr": "01_conversion.ipynb",
         "pad": "01_conversion.ipynb",
         "fill": "01_conversion.ipynb",
         "nbits": "01_conversion.ipynb",
         "num2ar": "01_conversion.ipynb",
         "ar2num": "01_conversion.ipynb",
         "ar2hex": "01_conversion.ipynb",
         "hex2ar": "01_conversion.ipynb",
         "str2ar": "01_conversion.ipynb",
         "ar2str": "01_conversion.ipynb",
         "COPY": "01_conversion.ipynb",
         "NOT": "01_conversion.ipynb",
         "AND": "01_conversion.ipynb",
         "OR": "01_conversion.ipynb",
         "Exclusive_OR": "01_conversion.ipynb",
         "XOR": "01_conversion.ipynb",
         "ar2gr": "01_conversion.ipynb",
         "gr2ar": "01_conversion.ipynb",
         "num2gr": "01_conversion.ipynb",
         "convert": "01_conversion.ipynb",
         "rint": "01_conversion.ipynb",
         "RNG": "01_conversion.ipynb",
         "depth": "02_recursion.ipynb",
         "flatten": "02_recursion.ipynb",
         "unflatten": "02_recursion.ipynb",
         "get": "02_recursion.ipynb",
         "give": "02_recursion.ipynb",
         "sort": "02_recursion.ipynb",
         "pipe": "02_recursion.ipynb",
         "maps": "02_recursion.ipynb",
         "replace": "03_templates.ipynb",
         "txt2lst": "03_templates.ipynb",
         "lst2txt": "03_templates.ipynb",
         "ZIP": "03_templates.ipynb",
         "getEvals": "03_templates.ipynb",
         "filltxt": "03_templates.ipynb",
         "Template": "03_templates.ipynb"}

modules = ["utils.py",
           "conversion.py",
           "recursion.py",
           "templates.py"]

doc_url = "https://Noeloikeau.github.io/sidis/"

git_url = "https://github.com/Noeloikeau/sidis/tree/master/"

def custom_doc_links(name): return None



__version__ = "0.0.1"

import time
import gzip
import pickle
import os
import typing
import numpy as np
from typing import Optional, Tuple, Dict, Callable, Union, Mapping, Sequence, Iterable
from functools import partial
import warnings
import collections

from .utils import *
from .conversion import *
from .recursion import *
from .templates import *


# AUTOGENERATED! DO NOT EDIT! File to edit: 03_templates.ipynb (unless otherwise specified).

__all__ = ['replace', 'txt2lst', 'lst2txt', 'ZIP', 'getEvals', 'filltxt', 'Template']

# Cell
from typing import Optional, Tuple, Dict, Callable, Union, Mapping, Sequence, Iterable
import numpy as np
from .conversion import cast
from .recursion import get

# Cell
def replace(l,**kwargs):
    'Replace the line `l` in `txt` with any occurances of the dictionaried values. '
    for k,v in kwargs.items():
        l=l.replace(k,str(v))
    return l

def txt2lst(txt):
    'Remove all newlines `\n` and empty strings `''` from the template `txt`.'
    return [t for t in txt.split('\n') if t!='']

def lst2txt(ltxt):
    'Concatenate a list of text `ltxt` into a string separated by newlines.'
    return '\n'.join(ltxt)

def ZIP(txt,_iter,*lambdas,as_txt=False):
    '''Loop over the iterable `_iter` and format `txt` in order of the `lambdas`.
    If `_iter` is an int, it is treated as `range(_iter)`, and if it is a tuple,
    it is treated as `np.ndindex(_iter)`.
    '''
    _iter=cast(_iter,list)
    try:
        l=[str(txt).format( *[l(*i) for l in lambdas] ) for i in _iter]
    except:
        l=[str(txt).format( *[l(i) for l in lambdas] ) for i in _iter]
    if as_txt is True:
        l=lst2txt(l)
    return l

def getEvals(replaced_txt,funcs=['ZIP']):
    '''Obtain a tuple of containing the `funcs`, the text they format, their arguments,
    and their line index in the template `txt`.
    '''
    y=[]
    for i,o in enumerate(replaced_txt):
        for f in funcs:
            if len(o.split(f))>1:
                #function, (txt to be formatted, func args), position in file
                y+=[(f,o.split(f),i)]
    return y

def filltxt(txt,funcs=['ZIP'],**kwargs):
    '''Take a template `txt`, replace all `kwargs` via `Replace`, then evaluate the `funcs`
       on the surrounding text using `GetEvals`.'''
    txt=txt2lst(txt)
    for i,l in enumerate(txt):
        txt[i]=replace(l,**kwargs)
    y=getEvals(txt,funcs)
    e=list(map(lambda s: eval(s[0]+'('+'{0:s[1][0]}[0]'+','+s[1][1]+')' ),y)) #evaluate the function
    u=[(i,e) for i,e in zip([y[i][-1] for i in range(len(y))],[e[i] for i in range(len(y))])] #position/evals
    t=[] #new txt
    for i in range(len(txt)):
        t+=[txt[i]]
        for j,e in u: #remove old lines
            if i==j:
                t.remove(txt[i])
                for q in e:
                    t+=[q]
    return t


class Template:
    '''
    Automates iteration over arbitrary Python functions embedded into blocks of text.

    Class attrs:

        funcs=[`ZIP`] (list of callable) : the keyword function to map over the text it's embedded in

        filler=`fill` (callable) : method of filling the text


    Inputs:

        `temp` (str): input string to be filled, or file name to be loaded

        `kwargs` (dict): characters to be replaced and evaluated


    Attrs:

        `plate` (list): line-by-line filling of `temp` based on `kwargs`

        `temp` (str): saved version of `temp` separated by newlines


    Methods:

        `load`: loads `fname` and fills based on `kwargs`.

        `out`: write/append `txt`, which defaults to filled data `str`, to `fname`.

        `fill`: fills the template by replacing `kwargs` and evaluating `funcs`.

    '''
    funcs=['ZIP']
    filler=filltxt
    def __init__(self,temp,**kwargs):
        self.__dict__.update({k:v for k,v in kwargs.items() if k!='self' and k!='kwargs'})
        self.temp=temp
        try:
            self.load(temp,**kwargs)
            print(f'Loaded {temp}.')
        except:
            self.temp=temp
            self.fill(**kwargs)

    def fill(self,temp=None,append=False,**kwargs):
        if not append:
            if temp is not None:
                self.temp=temp
            if kwargs!={}:
                self.plate=Template.filler(self.temp,Template.funcs,**kwargs)
                self.__dict__.update(kwargs)
        else:
            if (temp is not None) and (temp!=self.temp):
                self.temp+='\n'+temp
            if kwargs!={}:
                self.plate+=Template.filler(temp,Template.funcs,**kwargs)
                self.__dict__.update(kwargs)

    def txt(self):
        return lst2txt(self.plate)

    def load(self,fname,**kwargs):
        with open(fname,'r') as f:
            self.temp=f.read()
        self.fill(**kwargs)

    def out(self,fname,txt=None,s='a'):
        if txt is None:
            txt=self.txt()
        with open(fname,s) as f:
            f.write(txt)

    def __repr__(self):
        try:
            return self.txt()
        except:
            return self.temp


# AUTOGENERATED! DO NOT EDIT! File to edit: 01_conversion.ipynb (unless otherwise specified).

__all__ = ['data', 'trycast', 'nonitr2itr', 'itr2nonitr', 'itr2itr', 'Caster', 'isiter', 'cast', 'typestr', 'pad',
           'fill', 'nbits', 'num2ar', 'ar2num', 'ar2hex', 'hex2ar', 'str2ar', 'ar2str', 'COPY', 'NOT', 'AND', 'OR',
           'Exclusive_OR', 'XOR', 'ar2gr', 'gr2ar', 'num2gr', 'convert', 'rint', 'RNG']

# Cell
import typing
import numpy as np
from typing import Optional, Tuple, Dict, Callable, Union, Mapping, Sequence, Iterable
from functools import partial
import warnings

# Cell
data = Union[None,int,float,list,tuple,str,dict,set,np.ndarray]

# Cell
def trycast(obj : data, to : type) -> data:
    '''
    Attempts to typecast `obj` to datatype `to`
    using default type conversion. Fallback
    of more complex casting cases in this module.
    '''
    try:
        return to(obj)
    except:
        return typing.cast(to,obj)

# Cell
def nonitr2itr(obj : data, to : type) -> data:
    '''
    Wrap the noniterable `obj` into the iterable type `to`.
    '''
    if to is list:
        return [obj]
    elif to is tuple:
        return (obj,)
    elif to is str:
        return f'{obj}'
    elif to is dict:
        return {obj:obj}
    elif to is set:
        return {obj}
    elif to is np.ndarray:
        return np.array(obj)
    else:
        return trycast(obj,to)

# Cell
def itr2nonitr(obj : data, to : type) -> data:
    '''
    Treat each element of the iterable `obj` as the
    noniterable type `to`.
    '''
    t=type(obj)
    if t is dict:
        return {k:trycast(v,to) for k,v in obj.items()}
    else:
        return trycast([trycast(i,to) for i in obj],to)

# Cell
def itr2itr(obj : data, to : type) -> data:
    '''
    Treat the iterable `obj` as a new iterable of type `to`.
    Treats dictionaries as their items rather than default keys,
    and treats np.ndarray as the callable np.array(obj).
    '''
    if type(obj) is dict:
        return trycast(obj.items(),to)
    elif to is dict:
        return {i:o for i,o in enumerate(obj)}
    elif to is np.ndarray:
        return np.array(obj)
    else:
        return trycast(obj,to)

# Cell
isiter = lambda t: hasattr(t,'__iter__')

class Caster:
    '''
    Universal typecasting class with customizable ruleset.
    Stores the ruleset as a dictionary of dictionaries.
    The outer dictionary stores the type of the object to be converted.
    The inner dictionary stores all the types to convert into.
    The values are partially evaluated functions on the inner type,
    which get called by a class object and evaluated on the outer type.
    The ruleset can be updated and changed as needed.
    '''
    types=[None,int,float,list,tuple,str,dict,set,np.ndarray]

    iterables=[t for t in types if isiter(t)]

    iterables.remove(str) #want to wrap strings like numbers

    def get_rules(types=types,
                  iterables=iterables,
                  itr2itr=itr2itr,
                  nonitr2itr=nonitr2itr,
                  itr2nonitr=itr2nonitr):

        rules={t1:{t2:None for t2 in types} for t1 in types}
        noniterables=[t for t in types if t not in iterables]
        Caster.noniterables=noniterables
        for t1 in types:
            for t2 in types:
                if t1 in noniterables and t2 in iterables:
                    rules[t1][t2]=partial(nonitr2itr,to=t2)
                elif t1 in iterables and t2 in noniterables:
                    rules[t1][t2]=partial(itr2nonitr,to=t2)
                elif t1 in iterables and t2 in iterables:
                    rules[t1][t2]=partial(itr2itr,to=t2)
                else:
                    rules[t1][t2]=partial(trycast,to=t2)

        return rules


    def __init__(self,
                 rules : Optional[dict] = None):
        if rules is None:
            rules=Caster.get_rules()
        self.rules=rules

    def __getitem__(self,item):
        return self.rules[item]


    def __call__(self,
                 obj : data,
                 *args : type):
        res=obj
        for arg in args:
            t=type(res)
            try:
                res=self.rules[t][arg](res)
            except:
                res=trycast(res,arg)
        return res

# Cell
cast=Caster()

# Cell
def typestr(x : data):
    '''
    Parses the string of the input type for readability.
    '''
    if type(x) is type: #if passing type itself
        s=x
    else: #otherwise get type of obj
        s=type(x)
    return str(s).split('<')[-1].split('>')[0].split('class')[-1].split('\'')[1]

# Cell
def pad(data : Union[np.ndarray,list],
          bits : Optional[int] = None,
          to : Union[int,float] = int) -> np.ndarray:
    '''
    Pads an array with zeros, up to a length of `bits`.
    '''
    if bits is None:
        bits=0
    else:
        bits=bits-len(data)
    x=[0 for i in range(bits)]+list(data)
    x=np.array(x).astype(to)
    return x

# Cell
def fill(x : list,fillwith=np.NaN,mask=True):
    '''
    Turn uneven nested lists `x` into arrays `y` substituting
    missing entries using `fillwith` and optionally masking.
    '''
    length = max(map(len, x))
    y=np.array([xi+[fillwith]*(length-len(xi)) for xi in x])
    if mask:
        if np.isfinite(fillwith):
            y=np.ma.masked_equal(y,fillwith)
        else:
            y=np.ma.masked_invalid(y)
    return y

# Cell
def nbits(x : Union[int,float,list,np.ndarray]) -> int:
    '''
    Return the number of bits required to represent the input `x`.
    '''
    t=type(x)
    if (t is int) or (t is float):
        bits=np.ceil(np.log2(x+1)).astype(int) if x!=0 else 1
    elif (t is list) or (t is np.ndarray):
        bits=len(x)
    return bits

# Cell
def num2ar(x: Union[int,float],
            bits : Optional[int] = None,
            to : Union[int,float] = int) -> np.ndarray:
    '''
    Converts decimal number `x` to array `a` zero-padded with `bits`.
    '''
    bits=bits or nbits(x) #if None, give default
    form='0'+str(bits)+'b'
    binary=format(int(x),form)
    with warnings.catch_warnings(): #ignore numpy deprecation warning
        warnings.simplefilter("ignore")
        a=np.fromstring(binary,'u1')-ord('0')
    return cast(a,to)

# Cell
def ar2num(a : Union[list,np.ndarray],
            to : Union[int,float] = int):
    '''
    Converts array `a` to decimal number `x`.
    '''
    temp=str()
    a=np.array(a).astype(int)
    for c in a:
        temp+=str(c)
    x=int(temp,2)
    return cast(x,to)

# Cell
def ar2hex(a : Union[list,np.ndarray],
            bits : Optional[int] = None,
            prefix : bool = True) -> str:
    '''
    Converts binary array to hex string
    in:
        a (numpy array) : binary array to convert
    out:
        h (str) : hex conversion of a
    '''
    bits=bits or nbits(a)
    form='0'+str(int(np.log2(bits)))+'x'
    h=format(ar2num(a),form)
    if prefix:
        h='0x'+h
    return h

# Cell
def hex2ar(h : str,
            bits : Optional[int] = None,
            to : Union[int,float] = int) -> np.ndarray:
    '''
    Converts a hex string `h` into an array `a` padded with `bits` and elements of `astype`.
    '''
    x=int(h,16)
    a=num2ar(x,bits,to)
    return a

# Cell
def str2ar(s : str,
            to : Union[list,np.ndarray] = np.ndarray) -> Union[list,np.ndarray]:
    '''
    Converts an input string `s` into an array or list `a` as per `astype`.
    '''
    a=[int(i) for i in s]
    return cast(a,to)

# Cell
def ar2str(a : Union[list,np.ndarray], to : data = None) -> str:
    '''
    Converts an input array `a` into a string `s`.
    '''
    s=''.join([str(cast(i,to)) for i in a])
    return s

# Cell
def COPY(x : Union[int,float]) -> Union[int,float]:
    '''
    Simply returns `x`.
    '''
    return x

def NOT(x : Union[int,float]) -> Union[int,float]:
    '''
    Return conjugate of `x`.
    '''
    return 1-x

def AND(x : Union[int,float],
        y : Union[int,float]) -> Union[int,float]:
    '''
    Return logical AND of `x` and `y`.
    '''
    return x*y

def OR(x : Union[int,float],
       y : Union[int,float]) -> Union[int,float]:
    '''
    Return logical OR of `x` and `y`. See DeMorgan's Laws.
    '''
    return x+y-x*y

def Exclusive_OR(x : Union[int,float],
                 y : Union[int,float]) -> Union[int,float]:
    '''
    Return logical exclusive OR of `x` and `y`. See DeMorgan's Laws.
    '''
    return OR( AND( x , NOT(y) ) , AND ( NOT(x) , y) )

def XOR(*args : Union[int,float,list,np.ndarray]) -> Union[int,float]:
    '''
    Arbitrary input XOR using recursiveness.
    '''
    x=0
    for a in args:
        x=Exclusive_OR(x,a)
    return x

# Cell
def ar2gr(binary : Union[list,np.ndarray],
             to : Union[list,np.ndarray] = np.ndarray) -> Union[list,np.ndarray]:
    '''
    Converts an input binary array to graycode.
    '''
    binary = cast(binary,int)
    gray = []
    gray += [binary[0]]
    for i in range(1,len(binary)):
        gray += [XOR(binary[i - 1], binary[i])]
    return cast(cast(gray,int),to)

# Cell
def gr2ar(gray : Union[list,np.ndarray],
             to : Union[list,np.ndarray] = np.ndarray) -> Union[list,np.ndarray]:
    '''
    Converts a gray-code array into binary.
    '''
    binary = []
    binary += [gray[0]]
    for i in range(1, len(gray)):
        if (gray[i] == 0):
            binary += [binary[i - 1]]
        else:
            binary += [NOT(binary[i - 1])]
    return cast(cast(binary,int),to)


# Cell
def num2gr(x:int,to:data=None):
    '''
    Converts decimal number `x` to equivalent gray-code number.
    '''
    return int(ar2str(ar2gr(num2ar(x))),2)

# Cell
def convert(obj : Union[int,float,list,hex,str,np.ndarray],
            to : Union[int,float,list,hex,str,np.ndarray] = np.ndarray,
            bits : Optional[int] = None,
            astype : Union[int,float,list,np.ndarray] = int,
            gray : bool = False):
    '''
    Converts an input `obj` into an output of type `to`, padding with `bits`.
    Internally converts `obj` to np.ndarray with elements of dtype `astype`,
    before converting to the desired dtype `to`. If `gray`, first converts this
    binary array to gray-code. If input or output are `hex`, requires prefix of `0x`.

    Possible conversions:
        int -> float
        int -> str
        int -> list
        int -> array
        int -> hex

        str -> int
        str -> float
        str -> list
        str -> array
        str -> hex

        list -> arr
        list -> int
        list -> float
        list -> str
        list -> hex

        arr -> list
        arr -> int
        arr -> float
        arr -> str
        arr -> hex

        hex -> int
        hex -> float
        hex -> arr
        hex -> list
        hex -> str

    '''

    t=type(obj)
    #first convert to binary numpy array
    if (t is np.ndarray) or (t is list):
        x=obj
    elif (t is int) or (t is float):
        x=num2ar(obj,bits,astype)
    else:# t is str
        if obj[:2]=='0x': #obj is hex
            x=hex2ar(obj[2:],bits,astype)
        else:
            x=str2ar(obj)
    x=cast(pad(x,bits),astype)
    g=cast(ar2gr(x),astype)
    #convert
    if (to is np.ndarray):
        if gray:
            return g
        else:
            return x
    elif (to is list) or (to is set):
        if gray:
            return to(g)
        else:
            return to(x)
    elif (to is int) or (to is float) or (to is hex):
        if gray:
            return to(ar2num(g))
        else:
            return to(ar2num(x))
    else:# to is str
        if gray:
            return ar2str(g)
        else:
            return ar2str(x)

# Cell
def rint(x: Union[int,float,list,np.ndarray]) -> Union[int,np.ndarray]:
    '''
    Typecast rounding to np arrays.

    '''
    t=type(x)
    if (t is list) or (t is np.ndarray):
        return np.rint(x).astype(int)
    else:
        return round(x)

# Cell
class RNG:
    '''
    Globally stable random number generator. Initialized with fixed `seed`.
    Contains `normal` and `random` methods, each with an `absval` and `asint` argument,
    which convert to positive values and round to integers respectively.
    '''
    def __init__(self,seed : Optional[int] = 0):
        self.rng=np.random.default_rng(seed)

    def typecast(self,
                 y : Union[int,float],
                 absval : bool = False,
                 asint : bool = False) -> Union[int,float,np.ndarray]:
        if absval:
            y=abs(y)
        if asint:
            y=rint(y)
        return y

    def normal(self,
               x : Union[float,int] = 0,
               y : Union[float,int] = 0,
               shape : Optional[tuple] = None,
               absval : bool = False,
               asint : bool = False) -> Union[int,float,np.ndarray]:
        '''
        Draw from a Gaussian distribution with mean `x` and standard deviation `y`.
        If `shape` is not None, return a numpy array of draws.
        '''
        return self.typecast(y=self.rng.normal(loc=x,scale=y,size=shape),absval=absval,asint=asint)

    def random(self,
               x : Union[float,int] = 0,
               y : Union[float,int] = 0,
               shape : Optional[tuple] = None,
               absval : bool = False,
               asint : bool = False) -> Union[int,float,np.ndarray]:
        '''
        Draw from a uniform distribution in the interval [`x`,`y`].
        If `shape` is not None, return a numpy array of draws.
        '''
        return self.typecast((y-x)*self.rng.random(size=shape),absval=absval,asint=asint)